// -*- C++ -*-

#pragma once

#include "kernel.h"
#include "gpu_types.h"
#include "gpu-utils.h"


///
/// \addtogroup math
/// \{

#ifdef __CYGWIN__
namespace std
{
  using ::cbrt;
}
#endif



namespace goopax
{
  namespace impl
  {
    namespace interface
    {
      extern "C"
      {
#define MAKEFUNC1(FUNC)							\
	GOOPAX_DLL_EXPORT(3.0.0) errno_type goopax_E_math_ ## FUNC(gpu_writer_base_impl* r, gpu_writer_base_impl* a, Int MODE) noexcept;
	GOOPAX_INSTANTIATE_MATH(MAKEFUNC1)

#undef MAKEFUNC1


	GOOPAX_DLL_EXPORT(3.0.0) errno_type goopax_E_math_pow_int_uint_floating_point(gpu_writer_base_impl*, gpu_writer_base_impl*, Int, Uint) noexcept;
	
      }
    }


#define MAKEFUNC(FUNC)				\
      inline gpu_half FUNC(gpu_half a) { /** sin(x*PI) \ingroup(math) */ gpu_half ret; check_errno(interface::goopax_E_math_ ## FUNC(ret.impl.get(), a.impl.get(), 0)); return ret; } \
      inline gpu_float FUNC(gpu_float a) { /** sin(x*PI) \ingroup(math) */ gpu_float ret; check_errno(interface::goopax_E_math_ ## FUNC(ret.impl.get(), a.impl.get(), 0)); return ret; } \
      inline gpu_double FUNC(gpu_double a) { /** sin(x*PI) \ingroup(math) */ gpu_double ret; check_errno(interface::goopax_E_math_ ## FUNC(ret.impl.get(), a.impl.get(), 0)); return ret;} \
      template <class T, class TA> inline auto FUNC(checknan<T, TA> a) -> decltype(FUNC(unrange(a))) \
      { return FUNC(unrange(a)); }

#define MAKEFUNC_INT(FUNC)						\
      inline gpu_half FUNC(gpu_half a, Int MODE) { gpu_half ret; check_errno(interface::goopax_E_math_ ## FUNC(ret.impl.get(), a.impl.get(), MODE)); return ret; } \
      inline gpu_float FUNC(gpu_float a, Int MODE) { gpu_float ret; check_errno(interface::goopax_E_math_ ## FUNC(ret.impl.get(), a.impl.get(), MODE)); return ret; } \
      inline gpu_double FUNC(gpu_double a, Int MODE) { gpu_double ret; check_errno(interface::goopax_E_math_ ## FUNC(ret.impl.get(), a.impl.get(), MODE)); return ret;} \
      template <class T, class TA> inline auto FUNC(checknan<T, TA> a, Int MODE) -> decltype(FUNC(unrange(a), MODE)) \
      { return FUNC(unrange(a), MODE); }
      

      GOOPAX_INSTANTIATE_MATH_ARG(MAKEFUNC)
      GOOPAX_INSTANTIATE_MATH_ARG_INT(MAKEFUNC_INT)

      
#undef MAKEFUNC
#undef MAKEFUNC_INT


      
      template <typename T, typename=typename enable_if<is_floating_point<T>::value>::type>
      inline gpu_type<T> pow(gpu_type<T> a, Int ENUM, Uint DENOM=1)
      {
	/// Returns a to the power (ENUM/DENOM). \ingroup math
	gpu_type<T> ret;
	check_errno(interface::goopax_E_math_pow_int_uint_floating_point(ret.impl.get(), a.impl.get(), ENUM, DENOM));
	return ret;
      }

      template <Int ENUM, Uint DENOM, class T, class ENABLE=typename enable_if<is_floating_point<T>::value>::type>
      gpu_type<T> pow(const gpu_type<T> a)
      {
	/// Returns a to the power (ENUM/DENOM). \ingroup math
	return pow(a, ENUM, DENOM);
      }

      template <class T, class ENABLE=typename enable_if<is_floating_point<T>::value>::type>
      gpu_type<T> pow(const gpu_type<T> a, const gpu_type<T> b)
      {
	/// Returns a to the power b. \ingroup math
	return exp2(b * log2(a));
      }
      
      template <class T, class D, class ENABLE=typename enable_if<is_floating_point<T>::value && is_floating_point<D>::value>::type>
					    gpu_type<T> pow(const gpu_type<T> a, const D b)
      {
	/// Returns a to the power b. \ingroup math
	return pow(a, gpu_type<T>(b));
      }
      
      /// Square root \ingroup math
      inline gpu_half sqrt(gpu_half a) { return pow<1,2>(a); }
      /// Square root \ingroup math
      inline gpu_float sqrt(gpu_float a) { return pow<1,2>(a); }
      /// Square root \ingroup math
      inline gpu_double sqrt(gpu_double a) { return pow<1,2>(a); }
      /// Cubic root \ingroup math
      inline gpu_half cbrt(gpu_half a) { return pow<1,3>(a); }
      /// Cubic root \ingroup math
      inline gpu_float cbrt(gpu_float a) { return pow<1,3>(a); }
      /// Cubic root \ingroup math
      inline gpu_double cbrt(gpu_double a) { return pow<1,3>(a); }
      /// returns e^a \ingroup math
      inline gpu_half exp(gpu_half a) { return exp2(a*(1.0/impl::log(2.0))); }
      /// returns e^a \ingroup math
      inline gpu_float exp(gpu_float a) { return exp2(a*(1.0/impl::log(2.0))); }
      /// returns e^a \ingroup math
      inline gpu_double exp(gpu_double a) { return exp2(a*(1.0/impl::log(2.0))); }
      /// returns log_n(a) \ingroup math
      inline gpu_half log(gpu_half a) { return log2(a) * impl::log(2.0); }
      /// returns log_n(a) \ingroup math
      inline gpu_float log(gpu_float a) { return log2(a) * impl::log(2.0); }
      /// returns log_n(a) \ingroup math
      inline gpu_double log(gpu_double a) { return log2(a) * impl::log(2.0); }
      /// sin \ingroup math
      inline gpu_half sin(gpu_half a) { return sinpi(a * (1.0 / PI)); }
      /// sin \ingroup math
      inline gpu_float sin(gpu_float a) { return sinpi(a * (1.0 / PI)); }
      /// sin \ingroup math
      inline gpu_double sin(gpu_double a) { return sinpi(a * (1.0 / PI)); }
      /// cos \ingroup math
      inline gpu_half cos(gpu_half a) { return cospi(a * (1.0 / PI)); }
      /// cos \ingroup math
      inline gpu_float cos(gpu_float a) { return cospi(a * (1.0 / PI)); }
      /// cos \ingroup math
      inline gpu_double cos(gpu_double a) { return cospi(a * (1.0 / PI)); }
      /// tan \ingroup math
      inline gpu_half tan(gpu_half a) { return tanpi(a * (1.0 / PI)); }
      /// tan \ingroup math
      inline gpu_float tan(gpu_float a) { return tanpi(a * (1.0 / PI)); }
      /// tan \ingroup math
      inline gpu_double tan(gpu_double a) { return tanpi(a * (1.0 / PI)); }
      /// asin \ingroup math
      inline gpu_half asin(gpu_half a) { return asinpi(a) * PI; }
      /// asin \ingroup math
      inline gpu_float asin(gpu_float a) { return asinpi(a) * PI; }
      /// asin \ingroup math
      inline gpu_double asin(gpu_double a) { return asinpi(a) * PI; }
      /// acos \ingroup math
      inline gpu_half acos(gpu_half a) { return acospi(a) * PI; }
      /// acos \ingroup math
      inline gpu_float acos(gpu_float a) { return acospi(a) * PI; }
      /// acos \ingroup math
      inline gpu_double acos(gpu_double a) { return acospi(a) * PI; }
	/// atan \ingroup math
	inline gpu_half atan(gpu_half a) { return atanpi(a) * PI; }
	/// atan \ingroup math
	inline gpu_float atan(gpu_float a) { return atanpi(a) * PI; }
	/// atan \ingroup math
	inline gpu_double atan(gpu_double a) { return atanpi(a) * PI; }

  }
}
/// \}
