// -*- C++ -*-

#pragma once

#include "proto.h"
#include "lib/stdinc.h"
#include "lib/vec.h"
#include "lib/range.h"
#include "gpu_defs.h"
#include "cpu_functions.h"
#include "operators.h"

#define GOOPAX_CONSERVATIVE_ARITHMETIC_CONVERSIONS 0

/// \addtogroup types
/// gpu types, used in gpu kernels.
///

/// \addtogroup operators


namespace goopax
{

  template <typename T> struct goopax_struct_type<T, typename std::conditional<true, void, typename T::goopax_struct_type>::type>
  {
    using type = typename T::goopax_struct_type;
  };


    


  template <typename STR, typename X> struct goopax_struct_changetype<STR, X, typename std::conditional<true, void, typename STR::template goopax_struct_changetype<X> >::type>
  {
    using type = typename STR::template goopax_struct_changetype<X>;
  };

    
    
  namespace impl
  {

    struct xtype_int
    {
    };
    using xtype_gpu = gpu_type<Uint>;
    using xtype_void = xtype_int;

    struct gpu_writer_base_impl;
    struct kernel_impl;
    struct texbuf;
    struct loop_impl;
    struct resource_impl;
    struct gpu_if_impl;
    struct WELL512_impl;
    struct vecWELL512lib_impl;
    struct sha256_impl;
    struct perfdata_math_entry_impl;
    class svm_buffer_impl;
    namespace gpu_struct
    {
	struct mapframe_base;
      template <class STR> struct mapframe;
    }
      
    namespace interface
    {
#define MAKE_DELETE(T)							\
      extern "C" GOOPAX_DLL_EXPORT(3.0.0) void goopax_delete2_ ## T(void* p) noexcept;	\
      inline void goopax_delete(T* p)					\
      {									\
	goopax_delete2_ ## T(p);						\
      }

      MAKE_DELETE(gpu_writer_base_impl)
      MAKE_DELETE(kernel_impl)
      MAKE_DELETE(loop_impl)
      MAKE_DELETE(gpu_if_impl)
      MAKE_DELETE(WELL512_impl)
      MAKE_DELETE(vecWELL512lib_impl)
      MAKE_DELETE(sha256_impl)
      MAKE_DELETE(perfdata_math_entry_impl)
      MAKE_DELETE(svm_buffer_impl)

      extern "C" GOOPAX_DLL_EXPORT(3.0.0) void goopax_delete_resource_base2(void* p) noexcept;
      inline void goopax_delete(resource_impl* p)
      {
	goopax_delete_resource_base2(p);
      }

      
#undef MAKE_DELETE
#undef MAKE_INTRUSIVE

      struct goopax_deleter 
      {
	template <typename T>
	void operator()(T* p) 
	{
	  goopax_delete(p);
	}
      };

      struct goopax_nodelete_deleter 
      {
	template <typename T>
	void operator()(T*) 
	{
    }
      };


      
      extern "C"
      {
	GOOPAX_DLL_EXPORT(3.0.0) void goopax_gpu_type_op1(Tuint op, gpu_writer_base_impl* ret, const gpu_writer_base_impl* a);
	GOOPAX_DLL_EXPORT(3.0.0) void goopax_gpu_type_op2(Tuint op, gpu_writer_base_impl* ret, const gpu_writer_base_impl* a, const gpu_writer_base_impl* b);
	GOOPAX_DLL_EXPORT(3.0.0) void goopax_gpu_type_op3(Tuint op, gpu_writer_base_impl* ret, const gpu_writer_base_impl* a, const gpu_writer_base_impl* b, const gpu_writer_base_impl* c);
	GOOPAX_DLL_EXPORT(3.0.0) void goopax_gpu_type_comment2(gpu_writer_base_impl*, const char*) noexcept;
	GOOPAX_DLL_EXPORT(3.0.0) gpu_writer_base_impl* goopax_gpu_type_new_impl2() noexcept;
	GOOPAX_DEPRECATED GOOPAX_DLL_EXPORT(3.0.0) gpu_writer_base_impl* goopax_gpu_type_uninit2(Uint) noexcept;
	  GOOPAX_DLL_EXPORT(3.1.2) gpu_writer_base_impl* goopax_gpu_type_uninit3(Uint, memory::scope) noexcept;
	GOOPAX_DLL_EXPORT(3.0.0) gpu_writer_base_impl* goopax_gpu_type_assign2(const void* b, Uint) noexcept;
	GOOPAX_DLL_EXPORT(3.0.0) gpu_writer_base_impl* goopax_gpu_type_convert2(const gpu_writer_base_impl*, Uint) noexcept;
	GOOPAX_DLL_EXPORT(3.0.0) errno_type goopax_E_loop_break3(gpu_writer_base_impl*) noexcept;
	GOOPAX_DLL_EXPORT(3.1.2) errno_type goopax_E_loop_continue(gpu_writer_base_impl*) noexcept;
      }

      extern "C" GOOPAX_DLL_EXPORT(3.0.0) gpu_writer_base_impl* goopax_gpu_type_copyctor2(const gpu_writer_base_impl*) noexcept;
      extern "C" GOOPAX_DLL_EXPORT(3.0.0) void goopax_gpu_type_copyassign2(gpu_writer_base_impl*, const gpu_writer_base_impl*) noexcept;

    }




    template <typename T> struct is_gpu_type;
    template <typename T, typename=void> struct get_X_type;

/*
 enum memory::scope
      {
	  memory::thread=0,
	  memory::thread=1,
	  memory::device=2,
	  memory::none=3
      };
 */

    namespace gpu_struct
    {
      template <typename T> struct struct_item_map;
      template <typename T, Size_t N> struct struct_mod_test;
        template <typename T, Uint POW> struct struct_type_probe;
        template <typename T, typename TO> struct struct_change_type;
    }
    template <typename T, memory::scope SCOPE> class local_scope_mem;
    template <typename T, typename size_type, Bool CONST> class resource;
    template <typename T, typename size_type> class buffer;

    template <typename T> struct is_gpu_struct;

    template <typename T, typename U> struct is_gpu_struct_help :
      public false_type
    {};

    template <typename T> struct is_gpu_struct_help<T, typename enable_if<(is_gpu_struct<typename goopax_struct_type<T>::type>::value ||
									   is_arithmetic<typename goopax_struct_type<T>::type>::value ||
									   is_pointer<typename goopax_struct_type<T>::type>::value ||
									   is_gpu_type<typename goopax_struct_type<T>::type>::value)>::type> :
      public integral_constant<bool, !is_arithmetic<T>::value && !is_pointer<T>::value>
    {};
    



    template <typename T> struct is_gpu_struct :
    public is_gpu_struct_help<typename remove_cv<T>::type, void>
    {};
    
    template <typename T> struct is_gpu_struct<gpu_type<T>> :
      public false_type
    {};

    
    

#if !GOOPAX_DISTRIB
    void GOOPAX_DLL_EXPORT(3.0.0) goopax_debugcount_impl();
#endif
    
    inline void debugcount()
    {
#if !GOOPAX_DISTRIB
      goopax_debugcount_impl();
#endif
    }


    template<bool is_unsigned, bool is_signed, Uint size> struct type_enum_raw2;
    template<> struct type_enum_raw2<false, true, 8> : public integral_constant<Uint, 0>{};
    template<> struct type_enum_raw2<false, true, 4> : public integral_constant<Uint, 1>{};
    template<> struct type_enum_raw2<false, true, 2> : public integral_constant<Uint, 2>{};
    template<> struct type_enum_raw2<false, true, 1> : public integral_constant<Uint, 3>{};
    template<> struct type_enum_raw2<true, false, 8> : public integral_constant<Uint, 4>{};
    template<> struct type_enum_raw2<true, false, 4> : public integral_constant<Uint, 5>{};
    template<> struct type_enum_raw2<true, false, 2> : public integral_constant<Uint, 6>{};
    template<> struct type_enum_raw2<true, false, 1> : public integral_constant<Uint, 7>{};

    template <typename T> struct type_enum_raw :
      public type_enum_raw2<is_signed<T>::value, is_unsigned<T>::value, sizeof(T)>
    {};

    template<> struct type_enum_raw<bool> : public integral_constant<Uint, 8>{};
    template<> struct type_enum_raw<Double> : public integral_constant<Uint, 9>{};
    template<> struct type_enum_raw<Float> : public integral_constant<Uint, 10>{};
    template<> struct type_enum_raw<Half> : public integral_constant<Uint, 11>{};

    template <typename T> struct type_enum :
      public integral_constant<Uint, 2*type_enum_raw<T>::value>
    {};
    
    template <typename T, typename TA> struct type_enum<checknan<T, TA>>:
      public integral_constant<Uint, 2*type_enum_raw<T>::value + 1>
    {};
    
    template <typename T> struct type_enum<T*> :
      public integral_constant<Uint, 2*(type_enum_raw<T>::value + 12)>
    {};
      
    template <typename T, typename TA> struct type_enum<checknan<T, TA>*> :
      public integral_constant<Uint, 2*(type_enum_raw<T>::value + 12) + 1>
    {};
      

    
      template <Uint N> struct reverse_type_enum_raw;
      
      template<> struct reverse_type_enum_raw<0> { using type = Uint64_t; };
      template<> struct reverse_type_enum_raw<1> { using type = Uint32_t; };
      template<> struct reverse_type_enum_raw<2> { using type = Uint16_t; };
      template<> struct reverse_type_enum_raw<3> { using type = Uint8_t; };
      template<> struct reverse_type_enum_raw<4> { using type = Int64_t; };
      template<> struct reverse_type_enum_raw<5> { using type = Int32_t; };
      template<> struct reverse_type_enum_raw<6> { using type = Int16_t; };
      template<> struct reverse_type_enum_raw<7> { using type = Int8_t; };
      template<> struct reverse_type_enum_raw<8> { using type = Bool; };
      template<> struct reverse_type_enum_raw<9> { using type = Double; };
      template<> struct reverse_type_enum_raw<10> { using type = Float; };
      template<> struct reverse_type_enum_raw<11> { using type = Half; };
      
      template <Uint N> struct reverse_type_enum
      {
      private:
	using type1 = typename reverse_type_enum_raw<(N/2)%12>::type;
	using type2 = typename conditional<N%2==1, checknan<type1>, type1>::type;
      public:
	using type = typename conditional<((N/2)>=12), typename unrangetype<type2>::type*, type2>::type;
      };
      

    using paramtype_t = CTuint;
    using param_size_type = goopax::impl::Tuint;

      /// \ingroup types
    namespace typedefs
    {
      using gpu_double = gpu_type<Double>; ///< 64 bit floating point type
      using gpu_float = gpu_type<Float>;   ///< 32 bit floatint point type
      using gpu_half = gpu_type<Half>;     ///< 16 bit floatint point
      using gpu_bool = gpu_type<bool>;     ///< boolean
      using gpu_char = gpu_type<char>;     ///< character type
      using gpu_schar = gpu_type<Schar>;   ///< signed char
      using gpu_uchar = gpu_type<Uchar>;   ///< unsigned char
      using gpu_short = gpu_type<Short>;   ///< short integer, corresponding to CPU type 'short'
      using gpu_int = gpu_type<Int>;       ///< integer
      using gpu_long = gpu_type<Long>;     ///< long
      using gpu_longlong = gpu_type<Longlong>; ///< long long
      using gpu_ushort = gpu_type<Ushort>; ///< unsigned short
      using gpu_uint = gpu_type<Uint>;     ///< unsigned int
      using gpu_ulong = gpu_type<Ulong>;   ///< unsigned long
      using gpu_ulonglong = gpu_type<Ulonglong>; ///< unsigned long long
      using gpu_int64 = gpu_type<Int64_t>; ///< 64 bit integer
      using gpu_int32 = gpu_type<Int32_t>; ///< 32 bit integer
      using gpu_int16 = gpu_type<Int16_t>; ///< 16 bit integer
      using gpu_int8 = gpu_type<Int8_t>;   ///< 8 bit integer
      using gpu_uint64 = gpu_type<Uint64_t>; ///< 64 bit unsigned int
      using gpu_uint32 = gpu_type<Uint32_t>; ///< 32 bit unsigned int
      using gpu_uint16 = gpu_type<Uint16_t>; ///< 16 bit unsigned int
      using gpu_uint8 = gpu_type<Uint8_t>;  ///< 8 bit unsigned int
        using gpu_size_t = gpu_uint64;
        using gpu_ssize_t = gpu_int64;
    }
    using namespace typedefs;


    template <typename T> struct is_gpu_type :
      public false_type
    {
    };

    template <typename T> struct is_gpu_type<gpu_type<T>> :
      public true_type
    {};



    /**
     \ingroup types
     \brief type info and type conversion mechanism
       With the gettype struct, CPU and GPU types can be converted to each other, and type information obtained
       typename gettype<T>::cpu converts T to the corresponding CPU type. T can be CPU type, GPU type, or CPU debug type, or a goopax struct type.
       typename gettype<T>::gpu converts T to the corresponding GPU type. T can be CPU type, GPU type, or CPU debug type, or a goopax struct type.
       gettype<T>::size provides the size of type T in bytes. T can be CPU type, GPU type, or CPU debug type, or a goopax struct type.
       typename gettype<T>::template change<U>::type changes the type of T to type U while preserving CPU/GPU/debug mode. If T is CPU type, the resulting type is also a CPU type. If T is GPU type, the resulting type is a GPU type. If T is debug data type, the resulting type is also a debug data type.
     */
    template <typename T, typename EIF=void> struct gettype;


    template <typename T> struct gettype_base
    {
      static_assert(is_arithmetic<T>::value || is_pointer<T>::value, "");
      using cpu = goopax::impl::checknantype<T>;
      using gpu = gpu_type<T>;
      static GOOPAX_CONSTEXPR Size_t size = sizeof(T);
    };
#if __cplusplus < 201700
    template <typename T> GOOPAX_CONSTEXPR Size_t gettype_base<T>::size;
#endif

      template <typename T> struct gettype_base<T*>
      {
          using cpu = T*;
          using gpu = gpu_type<T*>;
	  static GOOPAX_CONSTEXPR Size_t size = sizeof(void*);
      };
#if __cplusplus < 201700
      template <typename T> GOOPAX_CONSTEXPR Size_t gettype_base<T*>::size;
#endif

      template <typename T> struct gettype_base<const T*>
      {
          using cpu = const T*;
          using gpu = gpu_type<const T*>;
          static GOOPAX_CONSTEXPR Size_t size = sizeof(void*);
      };
#if __cplusplus < 201700
      template <typename T> GOOPAX_CONSTEXPR Size_t gettype_base<const T*>::size;
#endif


    template <typename T> struct gettype<gpu_type<T>> :
      public gettype_base<T>
    {
      template <typename U> struct change
      {
	using type = gpu_type<typename unrangetype_core<U>::type>;
      };
    };
    
    template <typename T, typename TA> struct gettype<checknan<T, TA>> :
      public gettype_base<T>
    {
      template <typename U> struct change
      {
	using type = checknan<typename unrangetype_core<U>::type, TA>;
      };
    };
      
    template <typename TA> struct gettype<checknan<xtype_int, TA>>
    {
      template <typename U> struct change
      {
	using type = checknan<typename unrangetype_core<U>::type, TA>;
      };
    };
    template <> struct gettype<xtype_int>
    {
      template <typename U> struct change
      {
	using type = typename unrangetype_core<U>::type;
      };
    };
      

    template <typename T, Bool ARITH> struct gettype_split;
    
    template <typename STR> struct gettype_split<STR, false>
    {
      // defaults to goopax struct type
      using cpu = typename goopax_struct_changetype<typename remove_cv<STR>::type, goopax::impl::Tulong>::type;
      using gpu = typename goopax_struct_changetype<typename remove_cv<STR>::type, gpu_ulong>::type;
      static GOOPAX_CONSTEXPR Size_t size = sizeof(typename goopax_struct_changetype<typename remove_cv<STR>::type, Ulong>::type);
      template <typename U> struct change
      {
	using type = typename gettype<typename get_X_type<STR>::type>::template change<U>::type;
      };
    };
#if __cplusplus < 201700
    template <typename STR> GOOPAX_CONSTEXPR Size_t gettype_split<STR, false>::size;
#endif

    template <typename T> struct gettype_split<T, true> :
      public gettype_base<T>
    {
      template <typename U> struct change
      {
	using type = typename unrangetype<U>::type;
      };
    };


    
    template <typename T> struct gettype<T, typename enable_if<(is_arithmetic<T>::value || is_pointer<T>::value || is_gpu_struct<T>::value) && !is_checknan<T>::value>::type> :
      public gettype_split<T, is_arithmetic<T>::value || is_pointer<T>::value>
    {};

    

	// Sometimes getting alignment problems with std::unique_ptr. Using our own implementation here.
      template <typename IMPL> struct my_unique_ptr
      {
      private:
	IMPL* impl;
      public:

	IMPL* get() const
	{
	  return impl;
	}

	void reset(IMPL* ptr = nullptr)
	{
	  *this = my_unique_ptr(ptr);
	}

	IMPL& operator * () const
	{
	  return *get();
	}

	IMPL* operator -> () const
	{
	  return get();
	}
	
	my_unique_ptr(IMPL* ptr = nullptr) :
	  impl(ptr)
	{}
	my_unique_ptr(const my_unique_ptr&) = delete;
	my_unique_ptr& operator = (const my_unique_ptr&) = delete;
	my_unique_ptr(my_unique_ptr&& b) :
	  impl(b.impl)
	{
	  b.impl = nullptr;
	}
	my_unique_ptr& operator = (my_unique_ptr&& b)
	{
	  if (impl != nullptr)
	    {
	      goopax::impl::interface::goopax_deleter()(impl);
	    }
	  impl = b.impl;
	  b.impl = nullptr;
	  return *this;
	}
	~my_unique_ptr()
	{
	  if (impl != nullptr)
	    {
	      goopax::impl::interface::goopax_deleter()(impl);
	    }
	}
      };


      
      template <typename TYPE, typename IMPL> struct wrapper
      {						
	my_unique_ptr<IMPL> impl;
      };


    /// \ingroup operators
    /// conditional operator
    template <typename T> inline gpu_type<T> cond(const gpu_bool& c, const gpu_type<T>& a, const gpu_type<T>& b);
    

    struct INIT_IMPLNEW
    {
    };
    

    template <typename A, typename B, bool fA, bool fB, bool lt, bool is_signed> struct arith_result_type_help;
    
    template <typename A, typename B, bool lt, bool is_signed> struct arith_result_type_help<A, B, true, false, lt, is_signed>
    {
      using type = A;
    };      
    template <typename A, typename B, bool lt, bool is_signed> struct arith_result_type_help<A, B, false, true, lt, is_signed>
    {
      using type = B;
    };
    
    template <typename A, typename B> struct arith_result_type_help<A, B, true, true, true, true>
    {
      using type = B;
    };
    template <typename A, typename B> struct arith_result_type_help<A, B, true, true, false, true>
    {
      using type = A;
    };

    template <typename A, typename B> struct arith_result_type_help<A, B, false, false, true, false>
    {
      using type = B;
    };
    template <typename A, typename B> struct arith_result_type_help<A, B, false, false, false, false>
    {
      using type = A;
    };

    template <typename A, typename B, bool lt> struct arith_result_type_help<A, B, false, false, lt, true>
    {
      using type = typename make_signed<typename arith_result_type_help<A, B, false, false, lt, false>::type>::type;
    };
    

    template <typename A, typename B> struct arith_result_type :
      public arith_result_type_help<A, B, is_floating_point<A>::value, is_floating_point<B>::value, (gettype<A>::size < gettype<B>::size), (is_signed<A>::value || is_signed<B>::value)>
    {
      static_assert(!is_same<A, bool>::value, "");
      static_assert(!is_same<B, bool>::value, "");
    };
    


      /// \ingroup types
    /// GPU data type, used in GPU kernels to store values in registers.
    /// \tparam T corresponding CPU intrinsic type. Pointers are represented as gpu_type<T*> or gpu_type<const T*>, where T is a CPU intrinsic type or a goopax struct type.
      template <typename T, memory::scope scope> class gpu_type :
      public wrapper<gpu_type<T>, gpu_writer_base_impl>
    {
      static_assert(is_integral<T>::value || is_floating_point<T>::value, "");
      static_assert(!std::is_const<T>::value, "Please use non-const, non-volatile types.");
      static_assert(!std::is_volatile<T>::value, "Please use non-const, non-volatile types.");
	static_assert(scope==memory::thread, "scope must be thread for non-pointers");
	
      static GOOPAX_CONSTEXPR bool is_bool = is_same<T, bool>::value;
      static GOOPAX_CONSTEXPR bool is_int = is_integral<T>::value && !is_bool;
      

#define EIF(cond) template <typename EA=void, typename EB=typename enable_if<cond, EA>::type>

#define MAKE_OP0(PRE, RET, FUNC, ID)					\
      EIF(PRE) inline RET FUNC () const					\
	{								\
	  RET ret;							\
	  interface::goopax_gpu_type_op1(ID, ret.impl.get(), this->impl.get()); \
	  return ret;							\
	}
#define MAKE_OP1(PRE, RET, FUNC, ID, ARG1)			\
      EIF(PRE) inline RET FUNC (ARG1 b) const			\
	{								\
	  RET ret;							\
	  interface::goopax_gpu_type_op2(ID, ret.impl.get(), this->impl.get(), b.impl.get()); \
	  return ret;							\
	}
      
    public:
      /// Compare two GPU values for equality
      MAKE_OP1(true, gpu_bool, operator ==, 0, const gpu_type&)
      /// Compare two GPU values for inequality
	MAKE_OP1(true, gpu_bool, operator !=, 1, const gpu_type&)

      /// Compare two GPU values
	MAKE_OP1(!is_bool, gpu_bool, operator < , 2, const gpu_type&)
      /// Compare two GPU values
	MAKE_OP1(!is_bool, gpu_bool, operator >=, 3, const gpu_type&)
      /// boolean AND
	MAKE_OP1(is_bool, gpu_bool, operator &&, 4, const gpu_type&)
      /// boolean OR
	MAKE_OP1(is_bool, gpu_bool, operator ||, 5, const gpu_type&)
      /// addition
      MAKE_OP1(!is_bool, gpu_type, operator +, 6, const gpu_type&)
      /// multiplication
	MAKE_OP1(!is_bool, gpu_type, operator *, 7, const gpu_type&)
      /// division
	MAKE_OP1(!is_bool, gpu_type, operator /, 8, const gpu_type&)
      /// bitwise AND
      MAKE_OP1(is_int, gpu_type, operator &, 9, const gpu_type&)
      /// bitwise OR
	MAKE_OP1(is_int, gpu_type, operator |, 10, const gpu_type&)
      /// bitwise xor
	MAKE_OP1(is_int, gpu_type, operator ^, 11, const gpu_type&)
      /// left shift
	MAKE_OP1(!is_bool, gpu_type, operator <<, 12, const gpu_uint&)
      /// right shift
	MAKE_OP1(!is_bool, gpu_type, operator >>, 13, const gpu_uint&)
      /// modulo operator
	MAKE_OP1(is_int, gpu_type, operator %, 14, const gpu_type&)



#if GOOPAX_CONSERVATIVE_ARITHMETIC_CONVERSIONS
#define MAKESYM(OP) template <typename B, typename=typename enable_if<!is_same<T, B>::value>::type> gpu_type<decltype(*(T*)nullptr OP (*(B*)nullptr))> operator OP (const gpu_type<B>& b) const \
      {									\
	static_assert(!is_same<T, B>::value, "");			\
	using WORKTYPE = gpu_type<decltype(*(T*)nullptr + (*(B*)nullptr))>; \
	return static_cast<WORKTYPE>(*this) OP (static_cast<WORKTYPE>(b)); \
      }
#else
#define MAKESYM(OP) template <typename B, typename=typename enable_if<!is_same<T, B>::value>::type> gpu_type<typename arith_result_type<T, B>::type> operator OP (const gpu_type<B>& b) const \
      {									\
	static_assert(!is_same<T, B>::value, "");			\
	using WORKTYPE = gpu_type<typename arith_result_type<T, B>::type>; \
	return static_cast<WORKTYPE>(*this) OP (static_cast<WORKTYPE>(b)); \
      }
#endif

#define MAKESYMB(OP) template <typename B, typename=typename enable_if<!is_same<T, B>::value>::type> gpu_bool operator OP (const gpu_type<B>& b) const \
      {									\
	static_assert(!is_same<T, B>::value, "");			\
	using WORKTYPE = gpu_type<typename arith_result_type<T, B>::type>; \
	return static_cast<WORKTYPE>(*this) OP (static_cast<WORKTYPE>(b)); \
      }

      MAKESYM(+)
      MAKESYM(-)
      MAKESYM(*)
      MAKESYM(/)
      MAKESYM(&)
	MAKESYM(|)
	MAKESYM(^)

	MAKESYMB(<)
	MAKESYMB(<=)
	MAKESYMB(>)
	MAKESYMB(>=)
	MAKESYMB(==)
	MAKESYMB(!=)
      
#undef MAKESYM
#undef MAKESYMB

	
	MAKE_OP0(is_int, gpu_type, operator ~, 4)
	MAKE_OP0(is_bool, gpu_bool, operator !, 5)

	EIF(!is_bool) inline gpu_type operator - (const gpu_type& b) const
      {
	/// subtract
	    return *this + (-b);
	}
      EIF(!is_bool) inline gpu_type operator - () const
      {
	/// unary -
	  return *this * T(-1);
	}
      EIF(!is_bool) inline gpu_type operator + () const
      {
	/// unary +
	return *this;
      }
      EIF(!is_bool) inline gpu_bool operator > (const gpu_type& b) const
      {
	/// comparison
	  return b < *this;
	}
      EIF(!is_bool) inline gpu_bool operator <= (const gpu_type& b) const
	{
	  /// comparison
	  return b >= *this;
	}
	

      EIF(is_int) inline gpu_type& operator ++ ()
      {
	/// increment
	  return *this += 1;
	}
      EIF(is_int) inline gpu_type& operator -- ()
	{
	  /// decrement
	  return *this -= 1;
	}
      EIF(is_int) inline gpu_type operator ++ (Int)
	{
	  /// increment
	  gpu_type<T> ret = *this;
	  ++*this;
	  return ret;
	}
      EIF(is_int) inline gpu_type operator -- (Int)
	{
	  /// decrement
	  gpu_type<T> ret = *this;
	  --*this;
	  return ret;
	}
      friend inline gpu_type operator + (T a, const gpu_type& b) {return gpu_type(a) + b;}
      friend inline gpu_type operator - (T a, const gpu_type& b) {return gpu_type(a) - b;}
      friend inline gpu_type operator * (T a, const gpu_type& b) {return gpu_type(a) * b;}
      friend inline gpu_type operator / (T a, const gpu_type& b) {return gpu_type(a) / b;}
      friend inline gpu_bool operator < (T a, const gpu_type& b) {return gpu_type(a) < b;}
      friend inline gpu_bool operator <=(T a, const gpu_type& b) {return gpu_type(a) <=b;}
      friend inline gpu_bool operator > (T a, const gpu_type& b) {return gpu_type(a) > b;}
      friend inline gpu_bool operator >=(T a, const gpu_type& b) {return gpu_type(a) >=b;}
      friend inline gpu_bool operator ==(T a, const gpu_type& b) {return gpu_type(a) ==b;}
      friend inline gpu_bool operator !=(T a, const gpu_type& b) {return gpu_type(a) !=b;}
      friend inline gpu_type operator &&(T a, const gpu_type& b) {return gpu_type(a) &&b;}
      friend inline gpu_type operator ||(T a, const gpu_type& b) {return gpu_type(a) ||b;}
      friend inline gpu_type operator & (T a, const gpu_type& b) {return gpu_type(a) & b;}
      friend inline gpu_type operator | (T a, const gpu_type& b) {return gpu_type(a) | b;}
      friend inline gpu_type operator ^ (T a, const gpu_type& b) {return gpu_type(a) ^ b;}
      friend inline gpu_type operator % (T a, const gpu_type& b) {return gpu_type(a) % b;}
      
      /// addition
      EIF(!is_bool) inline gpu_type& operator += (const gpu_type& b) { return *this = *this + b; }
      /// subtraction
      EIF(!is_bool) inline gpu_type& operator -= (const gpu_type& b) { return *this = *this - b; }
      /// multiplication
      EIF(!is_bool) inline gpu_type& operator *= (const gpu_type& b) { return *this = *this * b; }
      /// division
      EIF(!is_bool) inline gpu_type& operator /= (const gpu_type& b) { return *this = *this / b; }
      /// bitwise AND
      EIF(is_int) inline gpu_type& operator &= (const gpu_type& b) { return *this = *this & b; }
      /// bitwise OR
      EIF(is_int) inline gpu_type& operator |= (const gpu_type& b) { return *this = *this | b; }
      /// bitwise XOR
      EIF(is_int) inline gpu_type& operator ^= (const gpu_type& b) { return *this = *this ^ b; }
      /// left shift
      EIF(is_int) inline gpu_type& operator <<= (const gpu_uint& b) { return *this = *this << b; }
      /// right shift
      EIF(is_int) inline gpu_type& operator >>= (const gpu_uint& b) { return *this = *this >> b; }
      /// modulo
      EIF(is_int) inline gpu_type& operator %= (const gpu_type& b) { return *this = *this % b; }

#undef EIF
#undef MAKE_OP0
#undef MAKE_OP1

      void gpu_break()
      {
	/// Break from loop. Only valid on loop variables. Calling gpu_break() on non-loop variables results in an error.
	check_errno(interface::goopax_E_loop_break3(this->impl.get()));
      }
      void gpu_continue()
      {
	/// Continue loop. Only valid on loop variables. Calling gpu_continue() on non-loop variables results in an error.
	check_errno(interface::goopax_E_loop_continue(this->impl.get()));
      }

      inline void comment(const string& s) const
      {
	interface::goopax_gpu_type_comment2(this->impl.get(), s.c_str());
      }
      
      inline gpu_type() :
	gpu_type(INIT_IMPLNEW(), interface::goopax_gpu_type_uninit3(type_enum<T>::value, scope))
      {
	/// default constructor for uninitialized variable.
      }
    
      template <typename U> explicit inline gpu_type(const struct wrapper<gpu_type<U>, gpu_writer_base_impl>& b) :
	gpu_type(INIT_IMPLNEW(), interface::goopax_gpu_type_convert2(b.impl.get(), type_enum<T>::value))
      {
	/// explicit type conversion
      }

      inline gpu_type(T d) :
	gpu_type(INIT_IMPLNEW(), interface::goopax_gpu_type_assign2(&d, type_enum<T>::value))
      {
	/// conversion from CPU type
      }

      template <typename U, typename=typename enable_if<is_half<T>::value && is_arithmetic<U>::value>::type>
	inline gpu_type(U d) :
	  gpu_type((T)d)
      {
	/// conversion from CPU type
      }
      
      template <typename TA, typename BT> inline gpu_type(const checknan<BT, TA>& d) :
	gpu_type()
      {
	/// conversion from CPU debug type
	if (is_initialized(d))
	  *this = unrange(d);
      }
      
      template <typename B, typename FOO = typename enable_if<(((is_int &&
	is_integral<B>::value && !is_same<B, bool>::value) ||
	(is_floating_point<T>::value &&
	is_floating_point<B>::value)) &&
        gettype<T>::size >= gettype<B>::size) ||
	(is_floating_point<T>::value && is_integral<B>::value)>::type>
	
      inline gpu_type(const gpu_type<B>& b) :
	gpu_type(INIT_IMPLNEW(), interface::goopax_gpu_type_convert2(b.impl.get(), type_enum<T>::value))
      {
	/// \brief Implicit type conversion
	/// The type conversion rules are stricter than the standard C/C++ type conversion rules to avoid unintended performance loss:
	/// Integer values are only implicitly converted to integers of at least the same size or to floating point numbers.
	/// floating point numbers are only implicitly converted to other floating point numbers of at least the same size.
	/// Other conversions need to be done explicitly, with static_cast, or C/C++ style cast.
      }

      inline gpu_type(const gpu_type& b) :
	gpu_type(INIT_IMPLNEW(), interface::goopax_gpu_type_copyctor2(b.impl.get()))
      {
	/// copy constructor
      }
      inline gpu_type& operator = (const gpu_type& b)
      {
	/// assignment operator
	interface::goopax_gpu_type_copyassign2(this->impl.get(), b.impl.get());	
	return *this;
      }

      inline gpu_type(INIT_IMPLNEW, gpu_writer_base_impl* nt)
      {
	this->impl.reset(nt);
      }
    };

      
      
    
    template <typename RET, typename A> RET make_op(Tuint ID, const A& a)
    {
      RET ret;
      interface::goopax_gpu_type_op1(ID, ret.impl.get(), a.impl.get());
      return ret;
    }
    template <typename RET, typename A, typename B> RET make_op(Tuint ID, const A& a, const B& b)
    {
      RET ret;
      interface::goopax_gpu_type_op2(ID, ret.impl.get(), a.impl.get(), b.impl.get());
      return ret;
    }
    template <typename RET, typename A, typename B, typename C> RET make_op(Tuint ID, const A& a, const B& b, const C& c)
    {
      RET ret;
      interface::goopax_gpu_type_op3(ID, ret.impl.get(), a.impl.get(), b.impl.get(), c.impl.get());
      return ret;
    }
      
      template <typename T, typename Enable=typename enable_if<is_floating_point<T>::value>::type> 
      inline gpu_bool isfinite(gpu_type<T> f)
      {
          /// \ingroup Math
	/// returns true if f is finite
	return make_op<gpu_bool>(0, f);
      }

      template <typename T, typename Enable=typename enable_if<is_floating_point<T>::value>::type> 
      inline gpu_bool isinf(gpu_type<T> f)
      {
          /// \ingroup Math
	/// returns true if f is infinite
	return make_op<gpu_bool>(1, f);
      }

      template <typename T, typename Enable=typename enable_if<is_floating_point<T>::value>::type> 
      inline gpu_bool isnan(gpu_type<T> f)
      {
          /// \ingroup Math
	/// returns true if f is NaN
	return make_op<gpu_bool>(2, f);
      }

      template <typename T, typename Enable=typename enable_if<is_floating_point<T>::value>::type> 
      inline gpu_bool isnormal(gpu_type<T> f)
      {
          /// \ingroup Math
	/// returns true if f is normal
	return make_op<gpu_bool>(3, f);
      }


      template <typename T, typename Enable> 
      gpu_type<T> min(gpu_type<T> a, gpu_type<T> b)
      {
          /// \ingroup functions
	/// minimum of both numbers
	return make_op<gpu_type<T>>(15, a, b);
      }
      
      template <typename T, typename Enable> 
      gpu_type<T> max(gpu_type<T> a, gpu_type<T> b)
      {
          /// \ingroup functions
	/// maximum of both numbers
	return make_op<gpu_type<T>>(16, a, b);
      }

      template <typename T> 
      gpu_type<T> clamp(gpu_type<T> a, typename enable_if<true, gpu_type<T>>::type b, typename enable_if<true, gpu_type<T>>::type c)
      {
	/// \ingroup functions
	/// clamping a to range (b,c). Similar to max(min(a, c), b)
	return make_op<gpu_type<T>>(1, a, b, c);
      }
      
      template <typename T> 
      gpu_type<T> clamp(gpu_type<T> a, gpu_type<T> b, gpu_type<T> c)
      {
	/// \ingroup functions
	/// clamping a to range (b,c). Similar to max(min(a, c), b)
	return make_op<gpu_type<T>>(1, a, b, c);
      }
      

      template <typename T> gpu_type<T> min(gpu_type<T> a, typename unrangetype_core<T>::type b)
      {
          /// \ingroup functions
	return min(a, gpu_type<T>(b));
      }
      template <typename T> gpu_type<T> min(typename unrangetype_core<T>::type a, gpu_type<T> b)
      {
          /// \ingroup functions
	return min(gpu_type<T>(a), b);
      }

      template <typename T> gpu_type<T> max(gpu_type<T> a, typename unrangetype_core<T>::type b)
      {
          /// \ingroup functions
	return max(a, gpu_type<T>(b));
      }
      template <typename T> gpu_type<T> max(typename unrangetype_core<T>::type a, gpu_type<T> b)
      {
          /// \ingroup functions
	return max(gpu_type<T>(a), b);
      }

      template <typename V1, typename V2, typename BOOL, typename=typename enable_if<is_container<V1>::value && is_container<V2>::value>::type>
	inline auto cond(const BOOL& c, const V1& a, const V2& b)
	-> typename changetype<V1, typename remove_reference<decltype(cond(c, a[0], b[0]))>::type>::type
      {
          /// \ingroup operators
	/// conditional function, overload for vectors and arrays
	typename changetype<V1, decltype(cond(c, a[0], b[0]))>::type ret = a;
	for (Tsize_t k=0; k<ret.size(); ++k)
	  {
	    ret[k] = cond(c, a[k], b[k]);
	  }
	return ret;
      }

      template <typename T> inline gpu_type<T> cond(const gpu_bool& c, const gpu_type<T>& a, const gpu_type<T>& b)
      {
          /// \ingroup operators
	/// conditional function. Returns c ? a : b
	return make_op<gpu_type<T>>(0, c, a, b);
      }
      template <typename T> gpu_type<T> cond(gpu_bool c, gpu_type<T> a, typename unrangetype_core<T>::type b)
      {
	return cond(c, a, gpu_type<T>(b));
      }
      template <typename T> gpu_type<T> cond(gpu_bool c, typename unrangetype_core<T>::type a, gpu_type<T> b)
      {
	return cond(c, gpu_type<T>(a), b);
      }
      template <typename T, typename=typename enable_if<is_arithmetic<T>::value>::type>
      typename gettype<T>::gpu cond(gpu_bool c, T a, T b)
      {
	return cond(c, typename gettype<T>::gpu(a), typename gettype<T>::gpu(b));
      }
      template <typename T, typename=typename enable_if<is_arithmetic<T>::value>::type> T cond(bool c, T a, T b)
      {
	return c ? a : b;
      }

      template <typename T, typename=typename enable_if<(is_integral<T>::value && !is_same<T, bool>::value)>::type> 
      gpu_type<typename make_unsigned<T>::type> abs(gpu_type<T> a)
      {
          /// \ingroup functions
	/// \brief returns absolute value.
	/// if a is an integer, the return type is unsigned.
	return make_op<gpu_type<typename make_unsigned<T>::type>>(6, a);
      }

      template <typename T, typename=typename enable_if<is_floating_point<T>::value>::type> 
      gpu_type<T> abs(gpu_type<T> a)
      {
          /// \ingroup functions
	/// returns absolute value
	return make_op<gpu_type<T>>(6, a);
      }

      template <typename T, typename=typename enable_if<is_floating_point<T>::value>::type> 
      gpu_type<T> floor(gpu_type<T> a)
      {
          /// \ingroup Math
	/// round to -inf
	return make_op<gpu_type<T>>(7, a);
      }

      template <typename T, typename=typename enable_if<is_floating_point<T>::value>::type> 
      gpu_type<T> ceil(gpu_type<T> a)
      {
          /// \ingroup Math
	/// round to +inf
	return make_op<gpu_type<T>>(8, a);
      }

      template <typename T, typename=typename enable_if<is_floating_point<T>::value>::type> 
      gpu_type<T> round(gpu_type<T> a)
      {
          /// \ingroup Math
	/// round to nearest
	return make_op<gpu_type<T>>(9, a);
      }


      template <typename T, typename Enable=typename enable_if<(is_integral<T>::value && !is_same<T, bool>::value)>::type>
      gpu_uint clz(gpu_type<T> a)
      {
          /// \ingroup functions
	/// Counts the number of leading zeros in the given value of integral type
	return make_op<gpu_uint>(10, a);
      }
      
      template <typename T, typename Enable=typename enable_if<(is_integral<T>::value && !is_same<T, bool>::value)>::type> 
      gpu_uint popcount(gpu_type<T> a)
      {
          /// \ingroup functions
          /// Counts the number of bits
	return make_op<gpu_uint>(11, a);
      }
      
      template <typename T, typename Enable=typename enable_if<is_integral<T>::value && !is_same<T, bool>::value>::type>
	gpu_type<T> rotl(gpu_type<T> a, gpu_int b)
      {
          /// \ingroup functions
          /// rotate left
	return make_op<gpu_type<T>>(17, a, b);
      }

      template <typename T, typename Enable=typename enable_if<is_integral<T>::value && !is_same<T, bool>::value>::type>
	gpu_type<T> rotr(gpu_type<T> a, gpu_int b)
      {
          /// \ingroup functions
          /// rotate right
	return rotl(a, -b);
      }

      
#define MAKE_ARITH(OP)							\
      template <typename A, typename B, typename=typename enable_if<(is_floating_point<A>::value && \
							       is_integral<B>::value)>::type> \
      inline typename gettype<A>::gpu operator OP (const A& a, const gpu_type<B>& b) \
      {									\
	return typename gettype<A>::gpu(a) OP typename gettype<A>::gpu(b); \
      }									\
      template <typename A, typename B, typename=typename enable_if<(is_integral<A>::value && \
							       is_floating_point<B>::value)>::type>	\
      inline typename gettype<B>::gpu operator OP (const gpu_type<A>& a, const B& b) \
      {									\
	return typename gettype<B>::gpu(a) OP typename gettype<B>::gpu(b); \
      }

#define MAKE_FUNC(FUNC)							\
      template <typename A, typename B, typename=typename enable_if<(is_floating_point<A>::value && \
							       is_integral<B>::value)>::type> \
      inline typename gettype<A>::gpu FUNC (const A& a, const gpu_type<B>& b) \
      {									\
	return FUNC(typename gettype<A>::gpu(a), b);			\
      }									\
      template <typename A, typename B, typename=typename enable_if<(is_integral<A>::value && \
							       is_floating_point<B>::value)>::type>	\
      inline typename gettype<B>::gpu FUNC(const gpu_type<A>& a, const B& b) \
      {									\
	return FUNC(a, typename gettype<B>::gpu(b));			\
      }

#define MAKE_FUNC3(FUNC, ARG1)						\
      template <typename A, typename B, typename=typename enable_if<(is_floating_point<A>::value && \
							       is_integral<B>::value)>::type> \
      inline typename gettype<A>::gpu FUNC (ARG1 c, const A& a, const gpu_type<B>& b) \
      {									\
	return FUNC(c, typename gettype<A>::gpu(a), b);			\
      }									\
      template <typename A, typename B, typename=typename enable_if<(is_integral<A>::value && \
							       is_floating_point<B>::value)>::type>	\
      inline typename gettype<B>::gpu FUNC(ARG1 c, const gpu_type<A>& a, const B& b) \
      {									\
	return FUNC(c, a, typename gettype<B>::gpu(b));			\
      }

      MAKE_ARITH(+)
      MAKE_ARITH(-)
	MAKE_ARITH(*)
	MAKE_ARITH(/)
	MAKE_FUNC(max)
	MAKE_FUNC(min)
      MAKE_FUNC3(cond, gpu_bool)
      
#undef MAKE_ARITH
#undef MAKE_FUNC
#undef MAKE_FUNC3

	template <typename A, typename B, typename ENABLE=typename enable_if<is_integral<A>::value>::type> gpu_type<typename unrangetype_core<A>::type> operator << (A a, gpu_type<B> b)
      {
	return gpu_type<typename unrangetype_core<A>::type>(a) << b;
      }
      template <typename A, typename B, typename ENABLE=typename enable_if<is_integral<A>::value>::type> gpu_type<typename unrangetype_core<A>::type> operator >> (A a, gpu_type<B> b)
      {
	return gpu_type<typename unrangetype_core<A>::type>(a) >> b;
      }
  

    namespace interface
    {
      extern "C"
      {
	GOOPAX_DLL_EXPORT(3.0.0) void goopax_thread_local_id2(gpu_writer_base_impl*) noexcept;
	GOOPAX_DLL_EXPORT(3.0.0) void goopax_thread_group_id2(gpu_writer_base_impl*) noexcept;

	GOOPAX_DLL_EXPORT(3.0.0) errno_type goopax_E_thread_local_size2(Uint*) noexcept;
	GOOPAX_DLL_EXPORT(3.0.0) errno_type goopax_E_thread_num_subthreads2(Uint*) noexcept;
	GOOPAX_DLL_EXPORT(3.0.0) errno_type goopax_E_thread_num_groups2(Uint*) noexcept;
	GOOPAX_DLL_EXPORT(3.0.0) errno_type goopax_E_thread_global_size2(Uint*) noexcept;
	GOOPAX_DLL_EXPORT(3.0.0) errno_type goopax_E_thread_max_local_mem2(Size_t*) noexcept;

        }
    }


    template <typename T, typename> struct get_X_type
    {
      using type = typename get_X_type<typename goopax_struct_type<typename remove_cv<T>::type>::type>::type;
    };

    template <typename T> struct get_X_type<T, typename enable_if<(is_arithmetic<T>::value || is_pointer<T>::value) && !is_checknan<T>::value>::type>
    {
      using type = xtype_int;
    };
    template <> struct get_X_type<xtype_int>
    {
      using type = xtype_int;
    };
    template <typename T, typename TA> struct get_X_type<checknan<T, TA>>
    {
      using type = checknan<xtype_int, goopax::impl::thread_access_none>;
    };
    template <typename T> struct get_X_type<gpu_type<T>>
    {
      using type = xtype_gpu;
    };
    template<> struct get_X_type<void>
    {
      using type = void;
    };
    template <typename T> struct get_X_type<gpu_struct::struct_item_map<T>>
    {
      using type = gpu_struct::struct_item_map<typename get_X_type<T>::type>;
    };
    template <typename T, Size_t N> struct get_X_type<gpu_struct::struct_mod_test<T, N>>
    {
      using type = gpu_struct::struct_mod_test<typename get_X_type<T>::type, N>;
    };
      template <typename T, Uint POW> struct get_X_type<gpu_struct::struct_type_probe<T, POW>>
      {
          using type = gpu_struct::struct_type_probe<typename get_X_type<T>::type, POW>;
      };
      template <typename T, typename TO> struct get_X_type<gpu_struct::struct_change_type<T, TO>>
      {
          using type = gpu_struct::struct_change_type<typename get_X_type<T>::type, TO>;
      };
  }
  using impl::cond;


  using impl::gettype;

  using impl::is_gpu_type;
  
  using namespace impl::typedefs;

  using impl::cond;
  using impl::clz;
  using impl::popcount;
  using impl::rotl;
  using impl::rotr;
  using impl::clamp;

  using impl::isnan;
  using impl::isinf;
  using impl::isfinite;
  using impl::isnormal;
  using impl::floor;
  using impl::ceil;
  using impl::round;
  using impl::min;
  using impl::max;
  using impl::abs;

}


namespace std
{
  /// numeric_limits overload for GPU types
  template <typename T> struct numeric_limits<goopax::gpu_type<T>> :
    std::numeric_limits<T>
  {};

  using goopax::isnan;
  using goopax::isinf;
  using goopax::isfinite;
  using goopax::isnormal;
  using goopax::floor;
  using goopax::ceil;
  using goopax::round;
  using goopax::min;
  using goopax::max;
  using goopax::abs;
#if __cplusplus >= 201700
  using goopax::clamp;
#endif
}
