// -*- C++ -*-

#pragma once

#include "memory.h"

namespace goopax
{
  namespace impl
  {
    struct resource_impl;
    using cost_t = Double;
 
    
    namespace interface
    {
      extern "C"
      {
	GOOPAX_DLL_EXPORT(3.0.0) cost_t goopax_cost_global_read_impl2(Uint burstsize_bytes) noexcept;
	GOOPAX_DLL_EXPORT(3.0.0) cost_t goopax_cost_global_write_impl2(Uint burstsize_bytes) noexcept;
	GOOPAX_DLL_EXPORT(3.0.0) cost_t goopax_cost_local_read_impl2(Uint burstsize_bytes) noexcept;
	GOOPAX_DLL_EXPORT(3.0.0) cost_t goopax_cost_local_write_impl2(Uint burstsize_bytes) noexcept;
 	GOOPAX_DLL_EXPORT(3.0.0) Uint goopax_cost_global_burstsize2(Uint measure) noexcept;
 	GOOPAX_DLL_EXPORT(3.0.0) Uint goopax_cost_local_burstsize2(Uint measure) noexcept;
	GOOPAX_DLL_EXPORT(3.0.0) cost_t goopax_cost_REGBYTE2() noexcept;
	GOOPAX_DLL_EXPORT(3.0.0) cost_t goopax_cost_LDS_MEMBYTE2() noexcept;
	GOOPAX_DLL_EXPORT(3.0.0) cost_t goopax_cost_global_ACK2() noexcept;

	GOOPAX_DLL_EXPORT(3.0.0) cost_t goopax_cost_matherr_double2(Double err) noexcept;
	GOOPAX_DLL_EXPORT(3.0.0) cost_t goopax_cost_matherr_float2(Double err) noexcept;
	GOOPAX_DLL_EXPORT(3.0.0) cost_t goopax_cost_matherr_half2(Double err) noexcept;

	GOOPAX_DEPRECATED GOOPAX_DLL_EXPORT(3.0.0) void goopax_mem_global_barrier2() noexcept;
	GOOPAX_DLL_EXPORT(3.1.0) void goopax_mem_global_barrier3(Uint) noexcept;

	GOOPAX_DEPRECATED GOOPAX_DLL_EXPORT(3.0.0) void goopax_mem_ref_wrap_lock2() noexcept;
	GOOPAX_DEPRECATED GOOPAX_DLL_EXPORT(3.0.0) void goopax_mem_ref_wrap_unlock2() noexcept;
	GOOPAX_DLL_EXPORT(3.0.0) void goopax_reftest2(gpu_writer_base_impl*) noexcept;
	
        GOOPAX_DEPRECATED GOOPAX_DLL_EXPORT(3.0.0) gpu_writer_base_impl* goopax_res_ref2(resource_impl*, gpu_writer_base_impl*) noexcept;
	GOOPAX_DLL_EXPORT(3.0.2) void goopax_res_ref3(resource_impl*, gpu_writer_base_impl*, gpu_writer_base_impl*) noexcept;
	GOOPAX_DLL_EXPORT(3.0.0) void goopax_mem_barrier3(resource_impl*) noexcept;
	GOOPAX_DLL_EXPORT(3.0.0) Size_t goopax_lds_size2(resource_impl*) noexcept;
	
	GOOPAX_DLL_EXPORT(3.0.0) errno_type goopax_E_uav_create2(resource_impl**, Uint, Uint) noexcept;
	GOOPAX_DLL_EXPORT(3.0.0) errno_type goopax_E_uav_create_fromtex2(resource_impl**, texbuf*, texbuf*, Uint, Uint) noexcept;
	GOOPAX_DLL_EXPORT(3.0.0) errno_type goopax_E_lds_create_fromsizescope2(resource_impl**, Size_t, memory::scope, Uint) noexcept;
      }

      extern "C" GOOPAX_DLL_EXPORT(3.0.0) cost_t goopax_cost_instruction2(Uint) noexcept;
    }


    /// \ingroup COST
    /// \addtogroup COST
    /// \{
    /// Penalties used for cost function minimization. By using these functions to estimate a cost value, and
    /// by using it to select the methods with minimal cost, kernels can be written that have a good performance on various devices.
    
    namespace COST
    {
      template <class T> inline goopax::impl::Tuint global_burstsize()
      {
	/// Preferred read/write width per thread on global memory \ingroup COST
	return interface::goopax_cost_global_burstsize2(gettype<T>::size);
      }
      template <class T> inline goopax::impl::Tuint local_burstsize()
      {
	/// Preferred read/write width per thread on local memory \ingroup COST
	return interface::goopax_cost_local_burstsize2(gettype<T>::size);
      }

      template <class T> inline cost_t instruction()
      {
	/// Cost for each instructon of type T. \ingroup COST
	return interface::goopax_cost_instruction2(type_enum<T>::value);
      }
      
      inline cost_t matherr(Double err, goopax::impl::Tdouble)
      {
	/// Cost for allowing mathematical error for double precision calculations \ingroup COST
	return interface::goopax_cost_matherr_double2(err);
      }
      inline cost_t matherr(Double err, goopax::impl::Tfloat)
      {
	/// Cost for allowing mathematical error for single precision floating point calculations \ingroup COST
	return interface::goopax_cost_matherr_float2(err);
      }
      inline cost_t matherr(Double err, goopax::impl::Thalf)
      {
	/// Cost for allowing mathematical error for half precision floating point calculations \ingroup COST
	return interface::goopax_cost_matherr_half2(err);
      }

      template <class T> inline cost_t global_read(goopax::impl::Tuint burst)
      {
	/// Cost for reading burst contiguous values of type T from global memory   \ingroup COST
	return interface::goopax_cost_global_read_impl2(gettype<T>::size * burst);
      }
      template <class T> inline cost_t global_write(goopax::impl::Tuint burst)
      {
	/// Cost for writing burst contiguous values of type T to global memory   \ingroup COST
	return interface::goopax_cost_global_write_impl2(gettype<T>::size * burst);
      }
      template <class T> inline cost_t local_read(goopax::impl::Tuint burst)
      {
	/// Cost for reading burst contiguous values of type T from local memory   \ingroup COST
	return interface::goopax_cost_local_read_impl2(gettype<T>::size * burst);
      }
      template <class T> inline cost_t local_write(goopax::impl::Tuint burst)
      {
	/// Cost for writing burst contiguous values of type T to local memory   \ingroup COST
	return interface::goopax_cost_local_write_impl2(gettype<T>::size * burst);
      }
      
      template <class T> cost_t REG()
      {
	/// Cost for allocating an additional register \ingroup COST
	return gettype<T>::size * interface::goopax_cost_REGBYTE2();
      }
      template <class T> cost_t local_memory()
      {
	/// Cost for allocating local memory \ingroup COST
	return interface::goopax_cost_LDS_MEMBYTE2() * gettype<T>::size;
      }
      inline cost_t global_ACK()
      {
	/// Cost for memory synchronization \ingroup COST
	return interface::goopax_cost_global_ACK2();
      }
    }
    /// \}

  
    

  GOOPAX_DEPRECATED inline void global_barrier()
    {
        /// Places workgroup-wide barriers on all memory resources. No synchronization takes place between workgroups.
        /// This function has been renamed, please use local_barrier() instead.
        /// \sa{struct::barrier}
      interface::goopax_mem_global_barrier3(local_size());
    }
  
      inline void local_barrier()
      {
          /// Places workgroup-wide barriers on all memory resources. No synchronization takes place between workgroups.
          /// \sa{struct::barrier}
	interface::goopax_mem_global_barrier3(local_size());
      }

    namespace experimental
    {
      inline void global_barrier(unsigned int width_group=num_groups())
      {
	/// Places global barriers on all memory resources.
	/// Use with care, this could potentially cause infinite loops if too many workgroups are executed in parallel!
	/// \sa{struct::barrier}
	/// \param width_group If provided, splits the workgroups into blocks of size width_group and only places barriers within the threads in each block. width_group can be any value between 1 and num_groups(). If num_groups() is not divisible by width_group, the last block is smaller.
          interface::goopax_mem_global_barrier3(width_group*local_size());
      }
    }

    
    template <class P> struct getreftypes;

    template <class T> struct getreftypes<T*>
    {
      typedef T type;
      typedef T& ref;
  
    };
    template <class T> struct getreftypes<const T*>
    {
      typedef T type;
      typedef const T& ref;
    };



    template <class T, typename CPU_SIZE_TYPE> struct resource_type_base
    {
      using value_type = typename gettype<T>::gpu;
      using reference = value_type&;
      using const_reference = const value_type&;
      using baseT = T;
      static_assert(is_same<CPU_SIZE_TYPE, typename gettype<CPU_SIZE_TYPE>::cpu>::value, "");
      using cpu_size_type = CPU_SIZE_TYPE;
      using cpu_ssize_type = typename make_signed<cpu_size_type>::type;
      using gpu_size_type = typename gettype<cpu_size_type>::gpu;
      using gpu_ssize_type = typename gettype<cpu_ssize_type>::gpu;
      using size_type = gpu_size_type;
      using ssize_type = gpu_ssize_type;
      
      unique_ptr<resource_impl, interface::goopax_deleter> impl;
    };

    template <class T, typename cpu_size_type> struct resource_with_barrier_type_base :
      public resource_type_base<T, cpu_size_type>
    {
      inline void barrier() const
      {
	interface::goopax_mem_barrier3(this->impl.get());
      }
    };

    
    template <class T, typename CPU_SIZE_TYPE> struct uav :
      public resource_with_barrier_type_base<T, CPU_SIZE_TYPE>
    {
      static_assert(!is_gpu_type<T>::value, "");
      using BASE = resource_with_barrier_type_base<T, CPU_SIZE_TYPE>;
      
      using value_type = typename BASE::value_type;
      using reference = typename BASE::reference;
      using const_reference = typename BASE::const_reference;
      using baseT = T;
      using buffer_type = texture<T, CPU_SIZE_TYPE>;
      using typename BASE::size_type;

      size_type size() const
      {
	size_type ret;
	interface::goopax_resource_info(this->impl.get(), ret.impl.get(), 0);
	return ret;
      }

      inline uav(uav&&) = default;
      inline uav& operator = (uav&&) = default;
      inline uav()
      {
	this->impl.reset(if_call(interface::goopax_E_uav_create2, type_enum<T>::value, gettype<CPU_SIZE_TYPE>::size/gettype<paramtype_t>::size));
      }
      inline uav(texbuf* tex)
      {
	this->impl.reset(if_call(interface::goopax_E_uav_create_fromtex2, tex, (texbuf*)nullptr, type_enum<T>::value, gettype<CPU_SIZE_TYPE>::size/gettype<paramtype_t>::size));
      }
      inline uav(texbuf* tex, texbuf* real)
      {
	this->impl.reset(if_call(interface::goopax_E_uav_create_fromtex2, tex, real, type_enum<T>::value));
	
      }
      inline ~uav() = default;
    };
  


    template <class T, memory::scope SCOPE> struct lds :
      public conditional<SCOPE == memory::threadgroup, resource_with_barrier_type_base<T, goopax::impl::Tuint>, resource_type_base<T, goopax::impl::Tuint> >::type
    {
      using BASE = typename conditional<SCOPE == memory::threadgroup, resource_with_barrier_type_base<T, goopax::impl::Tuint>, resource_type_base<T, goopax::impl::Tuint> >::type;
      using value_type = typename BASE::value_type;
      using reference = typename BASE::reference;
      using const_reference = typename BASE::const_reference;
      using baseT = T;
      using typename BASE::cpu_size_type;

      inline cpu_size_type size() const
      {
	return interface::goopax_lds_size2(this->impl.get());
      }
    
      inline lds(lds&&) = default;
      inline lds& operator = (lds&&) = default;
      inline lds(cpu_size_type size0)
      {
	this->impl.reset(if_call(interface::goopax_E_lds_create_fromsizescope2, (Size_t)size0, SCOPE, type_enum<T>::value));
      }
      inline ~lds() = default; // FIXME: Barrier?
      
    };

    


  }
  namespace experimental = impl::experimental;
  using impl::global_barrier;
  using impl::local_barrier;
  namespace COST = impl::COST;
  using impl::cost_t;

}


