// -*- C++ -*-

#pragma once

#include "defs.h"
#include "functions.h"
#include "vec.h"



namespace goopax
{

  namespace impl
  {
    struct PARAMBASE;

    namespace interface
    {
      extern "C"
      {
	using outfunc_type = const char*(*)(const PARAMBASE*);
	
	GOOPAX_DLL_EXPORT(3.0.0) errno_type goopax_E_PARAMBASE_register2(PARAMBASE* THIS, const char* name0, const char* description0, const char* comment, void(*readfunc)(PARAMBASE*, const char*), outfunc_type, bool optional0) noexcept;
	GOOPAX_DLL_EXPORT(3.0.0) void goopax_PARAMBASE_unregister2(const char* name) noexcept;
      }
    }
    

    template <class=void> void readfunc(PARAMBASE* param, const char* value);
    template <class=void> const char* outfunc(const PARAMBASE* param);


    struct PARAMBASE
    {
    private:
      const string name_i;
      const string description_i;
    public:
      Tbool seen;
      virtual void read(const char* s) = 0;
  
      const string& name() const
      {
	return name_i;
      }
      const string& description() const
      {
	return description_i;
      }
      virtual string comment() const = 0;
      virtual void out(ostream& s) const = 0;
  
      PARAMBASE(const string& name0, const string& description0, const string& comment, Tbool optional) :
	name_i(name0),
	description_i(description0)
      {
	check_errno(interface::goopax_E_PARAMBASE_register2(this, name0.c_str(), description0.c_str(), comment.c_str(), &readfunc, &outfunc, optional));
      }

      PARAMBASE(const PARAMBASE&) = delete;
      PARAMBASE& operator = (const PARAMBASE&) = delete;
  
      virtual ~PARAMBASE()
      {
	interface::goopax_PARAMBASE_unregister2(name_i.c_str());
      }
    };

    template <class> void readfunc(PARAMBASE* param, const char* value)
    {
      param->read(value);
    }
    template <class> const char* outfunc(const PARAMBASE* param)
    {
      static string ret;
      stringstream ss;
      param->out(ss);
      ret = ss.str();
      return ret.c_str();
    }

    template <class T, bool IS_CLASS> struct autobase;


    template <class T> struct autobase<T, true> :
      public T
    {
    };

    template <class T> struct autobase<T, false>
    {
    private:
      T data;
    public:
  
      operator const T&() const
      {
	return data;
      }
      operator T&()
      {
	return data;
      }
    };

  

    template <typename T> void param_readhelp(T& t, const char* s, true_type)
    {
      t = s;
    }
    template <typename T> void param_readhelp(T& t, const char* s, false_type)
    {
      istringstream ss(s);
      ss.exceptions(ios::badbit | ios::failbit);
      ss >> t;
    }    


    template <class T> struct PARAM :
      public PARAMBASE,
      public autobase<T, is_class<typename unrangetype_core<T>::type>::value>
    {
      virtual void read(const char* s)
      {
	param_readhelp<T>(*this, s, is_string<T>());
      }

      virtual string comment() const
      {
	return pretty_typename(typeid(T));
      }

      inline const T& operator ()() const
      {
	return *this;
      }
      PARAM(const string& name, const string& description = "", goopax::impl::Tbool optional = false) :
	PARAMBASE(name, description, pretty_typename(typeid(T)), optional)
      {
      }    


      PARAM& operator = (const T& t)
      {
	static_cast<T&>(*this) = t;
	return *this;
      }

      virtual ~PARAM()
      {}

      void out(ostream& s) const
      {
	s << static_cast<const T&>(*this);
      }
    };


    template <class T> struct PARAMOPT :
      public PARAM<T>
    {
      const T def;

      virtual string comment() const
      {
	stringstream ss;
	ss << def;
	return PARAM<T>::comment() + " [" + ss.str() + "]";
      }

      PARAMOPT& operator = (const T& t)
      {
	PARAM<T>::operator = (t);
	return *this;
      }

      PARAMOPT(const string& name, const T& def0, const string& description = "") :
	PARAM<T>(name, description, true),
	def(def0)
      {
	static_cast<T&>(*this) = def;
      }    


      virtual ~PARAMOPT()
      {}
    };

  }
  using impl::PARAMOPT;
  using impl::PARAM;
}

