// -*- C++ -*-

#pragma once

#include "gpu_types.h"
#include "lib/param.h"
#include "device.h"

struct _cl_mem;
struct _cl_device_id;


namespace cl
{
  class Event;
  class CommandQueue;
}


namespace goopax
{
    /// \ingroup image
/// Image access flags
    enum IMAGE_FLAGS
    {
      IMAGE_READ_WRITE = 1,    ///< read/write access
      IMAGE_READ_ONLY = 2,     ///< read-only access
      IMAGE_WRITE_ONLY = 4,    ///< write-only access
      IMAGE_WRITE_DISCARD = 8  ///< write-only, previous values are not preserved
    };
    
    /// \ingroup memory
    /// Buffer access flags
    enum BUFFER_FLAGS
    {
      BUFFER_READ_WRITE = 1,     ///< read/write access
      BUFFER_READ_ONLY = 2,      ///< read-only access
      BUFFER_WRITE_ONLY = 4,     ///< write-only access
      BUFFER_WRITE_DISCARD = 8   ///< write-only, previous values are not preserved
    };

  inline IMAGE_FLAGS operator | (IMAGE_FLAGS a, IMAGE_FLAGS b)
  {
    return (IMAGE_FLAGS)((int)a | (int)b);
  }
  inline BUFFER_FLAGS operator | (BUFFER_FLAGS a, BUFFER_FLAGS b)
  {
    return (BUFFER_FLAGS)((int)a | (int)b);
  }
    

  namespace impl
  {
    struct texbuf;
    struct gl_buffer;
      namespace gpu_struct
      {
          template <typename STR> struct type_probe;
          template <typename STR, typename TO> struct changetype;
      }

    namespace interface
    {
      extern "C"
      {
	GOOPAX_DLL_EXPORT(3.0.0) errno_type goopax_E_utils_get_hash2(texbuf& buf, Uint hash[8]) noexcept;
	GOOPAX_DLL_EXPORT(3.0.0) errno_type goopax_E_utils_get_hash_ignore_uninit2(texbuf& buf, Uint hash[8]) noexcept;
      }
    }


    template <typename T=void> struct init_caller
    {
      init_caller() {}
      template<class FUNC, typename... A> init_caller(FUNC func, A... a)
      {
	func(a...);
      }
    };


    struct cl_kernel;
    struct texture_base;


    namespace interface
    {
      extern "C"
      {
	GOOPAX_DLL_EXPORT(3.0.0) void goopax_tab_lock2(texbuf*) noexcept;
	GOOPAX_DLL_EXPORT(3.0.0) void goopax_tab_unlock2(texbuf*) noexcept;
	GOOPAX_DEPRECATED GOOPAX_DLL_EXPORT(3.0.0) Size_t goopax_tab_size2(const texbuf*) noexcept;
	GOOPAX_DLL_EXPORT(3.1.1) Size_t goopax_tab_size3(const texbuf*) noexcept;
	GOOPAX_DLL_EXPORT(3.0.0) goopax_device_impl* goopax_tab_get_device(const texbuf*) noexcept;
	GOOPAX_DEPRECATED GOOPAX_DLL_EXPORT(3.0.0) bool goopax_tab_getdc2(const texbuf*) noexcept;
	GOOPAX_DEPRECATED GOOPAX_DLL_EXPORT(3.0.0) void goopax_tab_setdc2(texbuf*, bool) noexcept;
	GOOPAX_DEPRECATED GOOPAX_DLL_EXPORT(3.0.0) errno_type goopax_E_tex_lock_map(texbuf*, void**, void**) noexcept;
	GOOPAX_DLL_EXPORT(3.1.1) errno_type goopax_E_tex_lock_map2(texbuf*, void**, Size_t, Size_t) noexcept;
          GOOPAX_DLL_EXPORT(3.0.0) errno_type goopax_E_tex_release_map(texbuf*) noexcept;

	GOOPAX_DLL_EXPORT(3.0.0) errno_type goopax_E_tex_create_re2(texbuf**, texbuf*, Uint) noexcept;
	GOOPAX_DEPRECATED GOOPAX_DLL_EXPORT(3.0.0) errno_type goopax_E_tex_create2(texbuf**, goopax_device_impl* device, void (*)(texbuf&, texbuf&), Size_t, BUFFER_FLAGS, Uint*, void*, Uint) noexcept;
	GOOPAX_DLL_EXPORT(3.1.1) errno_type goopax_E_tex_create3(texbuf**, goopax_device_impl* device, Size_t, BUFFER_FLAGS, Uint*, void*, Uint, const gpu_struct::mapframe_base*, const gpu_struct::mapframe_base*) noexcept;
#if GOOPAX_HAVE_OPENCL
	GOOPAX_DLL_EXPORT(3.0.0) errno_type goopax_E_tex_create_cl2(texbuf**, goopax_device_impl* device, void (*)(texbuf&, texbuf&), _cl_mem*, Uint) noexcept;
#endif
#if GOOPAX_HAVE_CUDA
	GOOPAX_DLL_EXPORT(3.0.0) errno_type goopax_E_tex_create_cuda(texbuf**, goopax_device_impl* device, void (*)(texbuf&, texbuf&), void*, Size_t, Uint) noexcept;
#endif

          GOOPAX_DLL_EXPORT(3.0.1) errno_type goopax_E_copy(texbuf*, texbuf*, Size_t, Size_t, Size_t) noexcept;
	GOOPAX_DLL_EXPORT(3.0.0) errno_type goopax_E_copy_to_host(texbuf*, void*, Size_t, Size_t) noexcept;
	GOOPAX_DLL_EXPORT(3.0.0) errno_type goopax_E_copy_from_host(texbuf*, const void*, Size_t, Size_t) noexcept;
          GOOPAX_DLL_EXPORT(3.0.1) errno_type goopax_E_fill_buffer(texbuf*, const void*, Size_t, Size_t, Size_t) noexcept;

	GOOPAX_DLL_EXPORT(3.0.0) void goopax_resource_info(const resource_impl*, gpu_writer_base_impl*, Uint) noexcept;
	GOOPAX_DLL_EXPORT(3.1.0) errno_type goopax_E_get_ptr(resource_impl*, gpu_writer_base_impl*) noexcept;
      }
    }

    // FIXME: remove, perhaps
    struct INITTEX_REINTERPRET
    {};
    struct INITTEX_FROMGL
    {};
    struct INITTEX_IMPL
    {};
    
      struct map_base
      {
          texbuf* buf = nullptr;

          map_base(const map_base&) = delete;
          map_base& operator = (const map_base&) = delete;
          
          map_base(map_base&& b)
          {
              swap(buf, b.buf);
          }
          map_base& operator = (map_base&& b)
          {
              if (buf != nullptr)
                  check_errno(interface::goopax_E_tex_release_map(buf));
              buf = b.buf;
              b.buf = nullptr;
              return *this;
          }
          
          map_base() = default;
          ~map_base()
          {
              if (buf != nullptr)
                  check_errno(interface::goopax_E_tex_release_map(buf));
          }
      };
      


    struct tex_arg_base
    {
      texbuf* buf = nullptr;
      inline goopax_device get_device() const
      {
	if (buf == nullptr)
	  return nullptr;
	else
	  return interface::goopax_tab_get_device(buf);
      }
      
      inline array<Tuint, 8> get_hash() const
      {
	array<Uint, 8> ret;
	check_errno(interface::goopax_E_utils_get_hash2(*buf, &ret[0]));
	return convert<array<Tuint, 8>>(ret);
      }
      inline array<Tuint, 8> get_hash_ignore_uninit() const
      {
	array<Uint, 8> ret;
	check_errno(interface::goopax_E_utils_get_hash_ignore_uninit2(*buf, &ret[0]));
	return convert<array<Tuint, 8>>(ret);
      }

      inline tex_arg_base(texbuf* buf0) :
	buf(buf0)
      {}

      inline tex_arg_base(tex_arg_base&& b) :
	buf(b.buf)
      {
	b.buf = nullptr;
      }
      inline tex_arg_base& operator = (tex_arg_base&& b)
      {
	interface::goopax_tab_unlock2(buf);
	buf = b.buf;
	b.buf = nullptr;
	return *this;
      }

      virtual ~tex_arg_base()
      {
	interface::goopax_tab_unlock2(buf);
      }
      tex_arg_base() = default;
      tex_arg_base(const tex_arg_base&) = delete;
      tex_arg_base& operator = (const tex_arg_base&) = delete;
    };


    struct texture_base :
      public tex_arg_base
    {
      inline Size_t bytesize() const { return interface::goopax_tab_size3(buf); }

      texture_base(texbuf* buf) :
	tex_arg_base(buf)
      {}
      texture_base() = default;
    };



    template <typename A, typename B> struct combine_size_type
    {
      using type = typename conditional<(gettype<typename A::size_type>::size < gettype<typename B::size_type>::size), typename B::size_type, typename A::size_type>::type;
    };
  


    template <class T, typename size_type> struct uav;


    template <class T, typename CPU_SIZE_TYPE>
    struct texture :
      public texture_base
    {
      using cpu_size_type = CPU_SIZE_TYPE;
      using cpu_ssize_type = typename make_signed<cpu_size_type>::type;
      using gpu_size_type = typename gettype<cpu_size_type>::gpu;
      using gpu_ssize_type = typename gettype<cpu_ssize_type>::gpu;
      using size_type = cpu_size_type;
      using ssize_type = cpu_ssize_type;
      
      using resource_type = uav<T, cpu_size_type>;

      inline texbuf* get_buf() const
      {
	return this->buf;
      }

      inline texture(goopax_device device, const Size_t size, BUFFER_FLAGS flags, Uint* GLres, void* host_ptr, const gpu_struct::mapframe_base* mf, const gpu_struct::mapframe_base* mfu) :
	texture_base(if_call(interface::goopax_E_tex_create3, device.get_impl(), size, flags, GLres, host_ptr, type_enum<T>::value, mf, mfu))
      {}
      
      inline texture(texbuf& base, INITTEX_REINTERPRET) :
	texture_base(if_call(interface::goopax_E_tex_create_re2, &base, type_enum<T>::value))
      {}

#if GOOPAX_HAVE_OPENCL
      inline texture(goopax_device device, _cl_mem* mem) :
	texture_base(if_call(interface::goopax_E_tex_create_cl2, device.get_impl(), nullptr, mem, type_enum<T>::value))
      {}
#endif
#if GOOPAX_HAVE_CUDA
      inline texture(goopax_device device, void* mem, Size_t size) :
	texture_base(if_call(interface::goopax_E_tex_create_cuda, device.get_impl(), nullptr, mem, size, type_enum<T>::value))
      {}
#endif
      

      texture() = default;
      texture(const texture&) = delete;
      texture& operator = (const texture&) = delete;
        inline texture(texture&&) = default;
        inline texture& operator = (texture&&) = default;
    };

  }
  using impl::tex_arg_base;
}
