// -*- C++ -*-

#pragma once

#include <type_traits>


namespace goopax
{
    struct memory
    {
	enum scope
	{
	    thread,
	    threadgroup,
	    device,
	    none
	};
    };

    
    namespace impl
    {
	template <typename T, memory::scope scope=(std::is_pointer<T>::value ? memory::device : memory::thread)> class gpu_type;
    }
}


#define GOOPAX_INSTANTIATE_MATH_ARG_INT(FUNC)	\
  FUNC(root)

#define GOOPAX_INSTANTIATE_MATH_ARG(FUNC)	\
  FUNC(sinpi)					\
  FUNC(cospi)					\
  FUNC(tanpi)					\
  FUNC(exp2)					\
  FUNC(log2)					\
  FUNC(asinpi)					\
  FUNC(acospi)					\
  FUNC(atanpi)					\
  FUNC(sinh)					\
  FUNC(cosh)					\
  FUNC(tanh)					\
  FUNC(asinh)					\
  FUNC(acosh)					\
  FUNC(atanh)					\
  FUNC(erf)					\
  FUNC(erfc)					\
  FUNC(tgamma)					\


#define GOOPAX_INSTANTIATE_MATH_WRAP(FUNC)	\
  FUNC(sqrt)					\
  FUNC(cbrt)					\
  FUNC(exp)					\
  FUNC(log)					\
  FUNC(sin)					\
  FUNC(cos)					\
  FUNC(tan)					\
  FUNC(asin)					\
  FUNC(acos)					\
  FUNC(atan)


#define GOOPAX_INSTANTIATE_MATH(FUNC)		\
  GOOPAX_INSTANTIATE_MATH_ARG(FUNC)		\
  GOOPAX_INSTANTIATE_MATH_ARG_INT(FUNC)




namespace half_float
{
  class half;
}

namespace goopax
{

  namespace impl
  {
    template <typename T, typename THREAD_ACCESS> class checknan;
#define MAKEFUNC(FUNC)							\
    template <typename T, typename TA> inline T FUNC(goopax::impl::checknan<T, TA> a);
    
    GOOPAX_INSTANTIATE_MATH_WRAP(MAKEFUNC)
#undef MAKEFUNC
  }
}

namespace goopax
{
  namespace impl
  {

#define MAKEFUNC(FUNC)							\
    inline goopax::impl::gpu_type<half_float::half> FUNC(goopax::impl::gpu_type<half_float::half> a); \
    inline goopax::impl::gpu_type<float> FUNC(goopax::impl::gpu_type<float> a); \
    inline goopax::impl::gpu_type<double> FUNC(goopax::impl::gpu_type<double> a);
    
    
    GOOPAX_INSTANTIATE_MATH(MAKEFUNC)
    GOOPAX_INSTANTIATE_MATH_WRAP(MAKEFUNC)
#undef MAKEFUNC
  }
}

namespace std
{
#define MAKEFUNC(FUNC) using goopax::impl::FUNC;
  
  GOOPAX_INSTANTIATE_MATH(MAKEFUNC)
  GOOPAX_INSTANTIATE_MATH_WRAP(MAKEFUNC)

#undef MAKEFUNC
}




namespace goopax
{
    
/*    namespace impl
  {
      
//    template <typename T> class gpu_type;
  //  template <typename T> class gpu_type<T*>;
    //template <typename T> class gpu_type<const T*>;
  }
*/
  
  using impl::gpu_type;
}
